#ifndef CONFIGPARSER_HPP_INCLUDED
#define CONFIGPARSER_HPP_INCLUDED

#include <map>
#include <string>
#include <vector>
#include <sstream>

template <class t>
std::string ConfigParser_toString(t a);
template<class t, class t2>
t2& ConfigParser_conv(t &a, t2 &b);
template<class t>
std::string& ConfigParser_conv(t& a, std::string& b);

// To hold a config value
// any POD can be assigned and returned
struct cfgItem : public std::string
{
    cfgItem() = default;
    // Assignment
    cfgItem(std::string val);
    cfgItem(bool val);
    template<class t>
    cfgItem(t val): std::string(ConfigParser_toString(val)) {}
    // Conversions
    template<class t>
    operator t ();
    template<class t>
    operator t () const;
    operator bool() const;
};


// struct that holds options under a configuration section
struct cfgSection : public std::map<std::string, cfgItem>
{
    // Does a value exist
    bool has(const std::string name)const;
    // Get a value
    cfgItem get(const std::string name) const;
    // nothrow variation
    cfgItem get(const std::string name, const std::nothrow_t& nothrow)const;
    // Set[/replace] value
    bool set(const std::string name, const cfgItem val, bool replace = true);
    // remove value
    bool remove(const std::string name);
    // remove value if it matches val
    short remove(const std::string name, const cfgItem val);
    // Get list of value names
    std::vector<std::string> list() const;
};

// Class to parse configuration files
class ConfigParser
{
public:
    ConfigParser(const char* file = 0, bool SaveOnDestruct = 0);
    ~ConfigParser();
    // Opens a configuration file
    bool open(const std::string file);
    // Saves the configuration
    bool save(const char* as = 0);
    // Indicates if a configuration file is open
    bool is_open() const;

    // Returns whether the config has the section name
    bool has(const std::string name) const;
    // Returns true if a option is set under a section
    bool has(const std::string sect, const std::string name) const;

    // Gets a list of sections in the config
    std::vector<std::string> get() const;

    // Get cfgSection by name
    cfgSection get(const std::string name) const;
    // nothrow variation (returns empty cfgSection)
    cfgSection get(const std::string section, const std::nothrow_t& nothrow) const;

    // Get a cfgItem by section and name
    cfgItem get(const std::string section, const std::string name) const;
    // nothrow variation (returns "")
    cfgItem get(const std::string section, const std::string name, const std::nothrow_t& nothrow)const;


    // Adds a section
    bool add(const std::string name);
    // Adds a string
    int set(const std::string section, const std::string name, const cfgItem val, bool replace = 1);

    // Remove an option
    int remove(const std::string section, const std::string name);
    // Removes a option if it matches val
    int remove(const std::string section, const std::string name, const cfgItem val);
    // Removes a section
    bool remove(const std::string name);

    // Change exit save state
    void saveOnDestruct(bool newval);

private:
    std::map<std::string, cfgSection> sections;
    bool isOpen;
    bool saveAtDestruct;
    std::string file_name;
};

// Template implementations

template <class t>
std::string ConfigParser_toString(t a)
{
    std::ostringstream b;
    b << a;
    return b.str();
}
template<class t, class t2>
t2& ConfigParser_conv(t &a, t2 &b)
{
    std::stringstream c;
    c << a;
    c >> b;
    return b;
}
template<class t>
std::string& ConfigParser_conv(t &a, std::string &b)
{
    std::ostringstream c;
    c << a;
    return b.assign(c.str());
}
template<class t>
cfgItem::operator t ()
{
    t a;
    return ConfigParser_conv(*this, a);
}
template<class t>
cfgItem::operator t () const
{
    t a;
    return ConfigParser_conv(*this, a);
}

#endif // CONFIGPARSER_HPP_INCLUDED
